import os
import re
from smtplib import SMTPException, SMTPRecipientsRefused, SMTPResponseException

import pytest

from iposi import mail, MailError, MailFailure


def test_either_plain_text_or_html_required():
    with pytest.raises(ValueError, match="plain"):
        mail(
            sender="john@example.com",
            recipients="jane@example.org",
            subject="Some Test",
        )


def test_sending_fails(mock_smtp):
    # We test the following:
    # - A SMTPRecipientsRefused exception with a string as error is raised.
    # - A SMTPRecipientsRefused exception with a bytes array as error is raised.
    # - A SMTPResponseException exception with a string as error is raised.
    # - A SMTPResponseException exception with a bytes array as error is raised.
    # - A SMTPException is raised.
    # - No exception is raised.
    recipients = [
        "jane@example.org",
        "cecily@example.com",
        "peter@example.com",
        "ada@example.org",
        "doris@example.com",
        "bob@example.org",
    ]
    mock_smtp.return_value.__enter__.return_value.sendmail.side_effect = [
        SMTPRecipientsRefused(
            recipients={recipients[0]: (42, b"The answer was wrong")}
        ),
        SMTPRecipientsRefused(
            recipients={recipients[1]: (43, "The answer was still wrong")}
        ),
        SMTPResponseException(501, b"Bad recipient address"),
        SMTPResponseException(501, "Really bad recipient address"),
        SMTPException("What might the error be?"),
        {},  # is used as return value
    ]

    with pytest.raises(MailError) as exc_info:
        mail(
            sender="john@example.com",
            recipients=recipients,
            subject="Some Test",
            plain="Some text",
        )
    recipient_failures = exc_info.value.recipient_failures
    assert len(recipient_failures) == 5

    # first failure
    error = recipient_failures[recipients[0]]
    assert error.smtp_code == 42
    assert error.message == "The answer was wrong"
    assert error.smtp_exception is None

    # second failure
    error = recipient_failures[recipients[1]]
    assert error.smtp_code == 43
    assert error.message == "The answer was still wrong"
    assert error.smtp_exception is None

    # third failure
    error = recipient_failures[recipients[2]]
    assert error.smtp_code == 501
    assert error.message == "Bad recipient address"
    assert isinstance(error.smtp_exception, SMTPResponseException)

    # fourth failure
    error = recipient_failures[recipients[3]]
    assert error.smtp_code == 501
    assert error.message == "Really bad recipient address"
    assert isinstance(error.smtp_exception, SMTPResponseException)

    # fifth failure
    error = recipient_failures[recipients[4]]
    assert error.smtp_code is None
    assert error.message == "What might the error be?"
    assert isinstance(error.smtp_exception, SMTPException)


def test_sending_succeeds(mock_smtp):
    mock_smtp.return_value.__enter__.return_value.sendmail.return_value = {}

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain="Some text",
    )
    assert True


def test_configured_host_and_server_are_used(mock_smtp):
    os.environ["IPOSI_HOST"] = "some.host"
    os.environ["IPOSI_PORT"] = "42"

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain="Some text",
    )
    mock_smtp.assert_called_with(host="some.host", port=42)


def test_tls_used(mock_smtp):
    os.environ["IPOSI_USE_TLS"] = "1"
    mock_starttls = mock_smtp.return_value.__enter__.return_value.starttls

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain="Some text",
    )
    mock_starttls.assert_called_once()


def test_tls_not_used(mock_smtp):
    os.environ["IPOSI_USE_TLS"] = "0"
    mock_starttls = mock_smtp.return_value.__enter__.return_value.starttls

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain="Some text",
    )
    mock_starttls.assert_not_called()


def test_authentication_used(mock_smtp):
    os.environ["IPOSI_USERNAME"] = "admin"
    os.environ["IPOSI_PASSWORD"] = "not_very_secret"
    mock_login = mock_smtp.return_value.__enter__.return_value.login

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain="Some text",
    )
    mock_login.assert_called_with("admin", "not_very_secret")


def test_authentication_not_used(mock_smtp):
    del os.environ["IPOSI_USERNAME"]
    del os.environ["IPOSI_PASSWORD"]
    mock_login = mock_smtp.return_value.__enter__.return_value.login

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain="Some text",
    )
    mock_login.assert_not_called()


def test_email_is_sent_from_sender_to_single_recipient_with_subject(mock_smtp):
    mock_sendmail = mock_smtp.return_value.__enter__.return_value.sendmail

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain="Some text",
    )
    msg = mock_sendmail.call_args.kwargs["msg"]
    assert "From: john@example.com" in msg
    assert "To: jane@example.org" in msg
    assert "Subject: Some Test" in msg


def test_email_is_sent_from_sender_to_multiple_recipients_with_subject(mock_smtp):
    mock_sendmail = mock_smtp.return_value.__enter__.return_value.sendmail
    sender = "Ben Foster <foster@example.com>"
    recipients = ["Jerry Miller <jerry@example.org>", "Sarah Baker <sarah@example.org>"]

    mail(
        sender=sender,
        recipients=recipients,
        subject="Testing Recipients",
        plain="Some text",
    )
    assert len(mock_sendmail.call_args_list) == len(recipients)
    for index, call_args in enumerate(mock_sendmail.call_args_list):
        msg = call_args.kwargs["msg"]
        assert f"From: {sender}" in msg
        assert f"To: {recipients[index]}" in msg
        assert "Subject: Testing Recipients" in msg


def test_plain_text_and_html(mock_smtp):
    mock_sendmail = mock_smtp.return_value.__enter__.return_value.sendmail
    plain = "Some plain text content"
    html = "<p>Some HTML content</p>"

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain=plain,
        html=html,
    )
    msg = mock_sendmail.call_args.kwargs["msg"]
    content_pattern = (
        f"Content-Type: multipart/alternative.*"
        f"Content-Type: text/plain.*{plain}.*"
        f"Content-Type: text/html.*{html}"
    )
    assert re.search(content_pattern, msg, flags=re.DOTALL)


def test_plain_text_only(mock_smtp):
    mock_sendmail = mock_smtp.return_value.__enter__.return_value.sendmail
    plain = "Some plain text content"

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        plain=plain,
    )
    msg = mock_sendmail.call_args.kwargs["msg"]
    assert "text/html" not in msg
    content_pattern = (
        f"Content-Type: multipart/alternative.*Content-Type: text/plain.*{plain}"
    )
    assert re.search(content_pattern, msg, flags=re.DOTALL)


def test_html_only(mock_smtp):
    mock_sendmail = mock_smtp.return_value.__enter__.return_value.sendmail
    html = "<p>Some HTML content"

    mail(
        sender="john@example.com",
        recipients="jane@example.org",
        subject="Some Test",
        html=html,
    )
    msg = mock_sendmail.call_args.kwargs["msg"]
    assert "text/plain" not in msg
    content_pattern = (
        f"Content-Type: multipart/alternative.*Content-Type: text/html.*{html}"
    )
    assert re.search(content_pattern, msg, flags=re.DOTALL)


def test_mail_error_string_representation():
    mail_error = MailError(
        recipient_failures={
            "alice@example.org": MailFailure(
                smtp_code=None, message="Connection failed", smtp_exception=None
            ),
            "bob": MailFailure(
                smtp_code=501,
                message="Invalid address",
                smtp_exception=SMTPException("Something is wrong"),
            ),
        }
    )

    assert "alice@example.org" in str(mail_error)
    assert "Connection failed" in str(mail_error)
    assert "bob" in str(mail_error)
    assert "501" in str(mail_error)
    assert "Invalid address" in str(mail_error)



import os

import pytest
from pydantic_core._pydantic_core import ValidationError

from iposi import mail
from iposi.settings import get_settings


def _mail():
    mail(
        sender="John Doe <joe@example.com>",
        recipients="Jane Miller <jane@example.com>",
        subject="Test",
        plain="Some plain text content.",
        html="<html>Some HTML content.</html>",
    )


def test_all_settings_present(mock_smtp):
    _mail()


def test_host_is_required():
    del os.environ["IPOSI_HOST"]
    with pytest.raises(ValidationError):
        _mail()


def test_username_and_password_are_optional(mock_smtp):
    del os.environ["IPOSI_PASSWORD"]
    del os.environ["IPOSI_USERNAME"]
    _mail()


def test_username_requires_password():
    del os.environ["IPOSI_PASSWORD"]
    with pytest.raises(ValueError, match="must also be set"):
        _mail()


def test_password_requires_username():
    del os.environ["IPOSI_USERNAME"]
    with pytest.raises(ValueError, match="must also be set"):
        _mail()


def test_port_defaults_to_587():
    del os.environ["IPOSI_PORT"]
    settings = get_settings()
    assert settings.iposi_port == 587


def test_use_tls_defaults_to_true():
    del os.environ["IPOSI_USE_TLS"]
    settings = get_settings()
    assert settings.iposi_use_tls



import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import (
    SMTPException,
    SMTP,
    SMTPResponseException,
    SMTPRecipientsRefused,
)
from typing import NamedTuple, Optional, Sequence, Union, Dict

from iposi.settings import get_settings


class MailFailure(NamedTuple):
    """
    A failure which occurred while trying to send an email to a recipient.

    Parameters
    ----------
    smtp_code:
        The SMTP (error) code, if there is one.
    message:
        The error message.
    smtp_exception:
        The SMTP exception related to this error, if there is one.
    """

    smtp_code: Optional[int]
    message: str
    smtp_exception: Optional[SMTPException]


class MailError(Exception):
    """
    An error which occurred while trying to send an email.

    The error contains a property recipient_failures, which is a dictionary of recipient email
    addresses and corresponding MailFailure instances.

    Parameters
    ----------
    recipient_failures:
        A dictionary of recipient addresses and corresponding errors.
    """

    def __init__(self, recipient_failures: Dict[str, MailFailure]):
        self._recipient_failures = recipient_failures

    @property
    def recipient_failures(self) -> Dict[str, MailFailure]:
        return self._recipient_failures

    def __str__(self) -> str:
        return str(self.recipient_failures)


def mail(
    sender: str,
    recipients: Union[str, Sequence[str]],
    subject: str,
    plain: Optional[str] = None,
    html: Optional[str] = None,
) -> None:
    """
    Send an email.

    The email content can be provided as plain text or HTML or both. If neither plain
    text nor HTML content is provided, a ValueError is raised. Plain text or HTML is
    only included in the email if you provide it; no "auto-conversion" is attempted. A
    ValueError is raised if you provide neither plain text nor HTML.

    Some environment variables must be set for using this function; see the
    documentation for details.

    The sender and recipients can be specified as an email address or as a name
    followed by the email address in angular brackets (such as "John Doe
    <john@example.com>"). A string (for a single recipient) or a list of strings (for
    multiple recipients) may be passed as recipients.

    If you specify multiple recipients, the email is sent individually to each of the
    recipients, so that the To field of the email contains the address of the recipient
    only.

    The email is sent as a multipart message.

    Sending the email is attempted for all recipients, even if it has already failed
    for some of them. In case any of the attempts has failed, a MailError is raised.
    This exception contains a recipient_failures property, which is a dictionary. The
    keys of this dictionary are the recipient email addresses which failed, and the
    values are the corresponding MailFailure instances. Note this behaviour differs
    from that of the sendmail method of the standard SMTP class.

    This function is not intended for sending bulk emails. If you have more than a
    handful of recipients, you might consider using another solution.

    Parameters
    ----------
    sender:
        The sender, i.e. the email address for the email's From field.
    recipients:
        The recipients, i.e. the email addresses for the email's To field.
    subject:
        The content of the email's Subject field.
    plain:
        The email content as plain text.
    html:
        The email content as HTML.

    Raises
    ------
    MailError:
        If the email could not be sent for one or multiple recipients.

    """

    # Get the SMTP configuration
    # We do this first so that unit tests for the settings don't need to mock SMTP.
    settings = get_settings()

    # Check there is content to send
    if not plain and not html:
        raise ValueError("At least one of plain text or HTML must be provided.")

    # If a string is passed as recipients, convert it into a list
    if isinstance(recipients, str):
        recipients = [recipients]

    with smtplib.SMTP(host=settings.iposi_host, port=settings.iposi_port) as smtp:
        # Use TLS, if requested
        if settings.iposi_use_tls:
            smtp.starttls()

        # Login, if requested
        if settings.iposi_username and settings.iposi_password:
            smtp.login(settings.iposi_username, settings.iposi_password)

        # Send the mail to all recipients, one by one, collecting any errors along the
        # way
        recipient_failures: Dict[str, MailFailure] = {}
        for recipient in recipients:
            try:
                result = _send_single_mail(
                    sender=sender,
                    recipient=recipient,
                    subject=subject,
                    plain=plain,
                    html=html,
                    smtp=smtp,
                )
                # As we call the sendmail of the SMTP class with a single recipient,
                # we should never get a non-empty dictionary as result. However, we
                # nonetheless handle this case, in case a future change leads to calls
                # to sendmail with multiple recipients.
                for r, e in result.items():  # pragma: no cover
                    recipient_failures[r] = MailFailure(
                        smtp_code=e[0], message=str(e[1]), smtp_exception=None
                    )
            except SMTPRecipientsRefused as e:
                rec_smtp_code, rec_smtp_error = e.recipients[recipient]
                # A SMTPRecipientsRefused is essentially a container, so we don't
                # include it in our error
                rec_smtp_error_as_str = (
                    rec_smtp_error.decode()
                    if isinstance(rec_smtp_error, bytes)
                    else rec_smtp_error
                )
                recipient_failures[recipient] = MailFailure(
                    rec_smtp_code, rec_smtp_error_as_str, None
                )
            except SMTPResponseException as e:
                res_smtp_code = e.smtp_code
                res_smtp_error = e.smtp_error
                res_smtp_error_as_str = (
                    res_smtp_error.decode()
                    if isinstance(res_smtp_error, bytes)
                    else res_smtp_error
                )
                recipient_failures[recipient] = MailFailure(
                    res_smtp_code, res_smtp_error_as_str, e
                )
            except SMTPException as e:
                recipient_failures[recipient] = MailFailure(None, str(e), e)

        if len(recipient_failures) > 0:
            raise MailError(recipient_failures=recipient_failures)


def _send_single_mail(
    sender: str,
    recipient: str,
    subject: str,
    plain: Optional[str],
    html: Optional[str],
    smtp: SMTP,
) -> dict[str, tuple[int, bytes]]:
    # Construct the email
    message = MIMEMultipart("alternative")
    message["From"] = sender
    message["To"] = recipient
    message["Subject"] = subject
    if plain:
        message.attach(MIMEText(plain))
    if html:
        message.attach(MIMEText(html, "html"))

    # Send the email
    return smtp.sendmail(
        from_addr=sender, to_addrs=[recipient], msg=message.as_string()
    )



from typing import Optional

from pydantic import model_validator
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """
    Settings provided as environment variables.

    IPOSI_HOST: Hostname of the SMTP server
    IPOSI_PASSWORD: Password for the SMTP server
    IPOSI_PORT: Port on which the SMTP server is listening
    IPOSI_USERNAME: Username for the SMTP server
    IPOSI_USE_TLS: Whether to use TLS

    IPOSI_PASSWORD and IPOSI_USERNAME require each other.
    """

    iposi_host: str
    iposi_password: Optional[str] = None
    iposi_port: int = 587
    iposi_username: Optional[str] = None
    iposi_use_tls: bool = True

    @model_validator(mode="after")
    def username_requires_password(self) -> "Settings":
        if self.iposi_username and not self.iposi_password:
            raise ValueError(
                "The environment variable IPOSI_PASSWORD must also be set "
                "if IPOSI_USERNAME is set."
            )
        return self

    @model_validator(mode="after")
    def password_requires_username(self) -> "Settings":
        if self.iposi_password and not self.iposi_username:
            raise ValueError(
                "The environment variable IPOSI_USERNAME must also be set "
                "if IPOSI_PASSWORD is set."
            )
        return self


def get_settings() -> Settings:
    """
    Return the configuration settings.
    """
    return Settings()  # type: ignore
